@@ -7,7 +7,7 @@
  *
  * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
+ * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful, but
@@ -353,7 +353,7 @@ add_extension (ASN1_TYPE asn, const char *root, const char *extension_id,
   else
     snprintf (name, sizeof (name), "?LAST.extnValue");
 
-  result = _gnutls_x509_write_value (asn, name, ext_data, 0);
+  result = _gnutls_x509_write_value (asn, name, ext_data);
   if (result < 0)
     {
       gnutls_assert ();
@@ -397,7 +397,7 @@ overwrite_extension (ASN1_TYPE asn, const char *root, unsigned int indx,
   _gnutls_str_cpy (name2, sizeof (name2), name);
   _gnutls_str_cat (name2, sizeof (name2), ".extnValue");
 
-  result = _gnutls_x509_write_value (asn, name2, ext_data, 0);
+  result = _gnutls_x509_write_value (asn, name2, ext_data);
   if (result < 0)
     {
       gnutls_assert ();
@@ -1315,7 +1315,7 @@ _gnutls_x509_ext_gen_proxyCertInfo (int pathLenConstraint,
   if (pathLenConstraint < 0)
     {
       result = asn1_write_value (ext, "pCPathLenConstraint", NULL, 0);
-      if (result < 0)
+      if (result != ASN1_SUCCESS)
         result = _gnutls_asn2err (result);
     }
   else
@@ -1330,7 +1330,7 @@ _gnutls_x509_ext_gen_proxyCertInfo (int pathLenConstraint,
 
   result = asn1_write_value (ext, "proxyPolicy.policyLanguage",
                              policyLanguage, 1);
-  if (result < 0)
+  if (result != ASN1_SUCCESS)
     {
       gnutls_assert ();
       asn1_delete_structure (&ext);
@@ -1339,7 +1339,7 @@ _gnutls_x509_ext_gen_proxyCertInfo (int pathLenConstraint,
 
   result = asn1_write_value (ext, "proxyPolicy.policy",
                              policy, sizeof_policy);
-  if (result < 0)
+  if (result != ASN1_SUCCESS)
     {
       gnutls_assert ();
       asn1_delete_structure (&ext);
