@@ -55,6 +55,9 @@ struct node_st {
 struct gnutls_x509_trust_list_st {
   unsigned int size;
   struct node_st *node;
+
+  gnutls_x509_crt_t *blacklisted;
+  unsigned int blacklisted_size;
 };
 
 #define DEFAULT_SIZE 127
@@ -115,6 +118,11 @@ gnutls_x509_trust_list_deinit(gnutls_x509_trust_list_t list,
     if (!list)
         return;
 
+    for (j = 0; j < list->blacklisted_size; j++) {
+		gnutls_x509_crt_deinit(list->blacklisted[j]);
+    }
+    gnutls_free(list->blacklisted);
+
     for (i = 0; i < list->size; i++) {
         if (all)
             for (j = 0; j < list->node[i].trusted_ca_size; j++) {
@@ -159,21 +167,13 @@ gnutls_x509_trust_list_add_cas(gnutls_x509_trust_list_t list,
                                const gnutls_x509_crt_t * clist,
                                int clist_size, unsigned int flags)
 {
-    gnutls_datum_t dn;
-    int ret, i;
+    int i;
     uint32_t hash;
 
     for (i = 0; i < clist_size; i++) {
-        ret = gnutls_x509_crt_get_raw_dn(clist[i], &dn);
-        if (ret < 0) {
-            gnutls_assert();
-            return i;
-        }
-
-        hash = hash_pjw_bare(dn.data, dn.size);
+        hash = hash_pjw_bare(clist[i]->raw_dn.data, clist[i]->raw_dn.size);
         hash %= list->size;
 
-        _gnutls_free_datum(&dn);
         list->node[hash].trusted_cas =
             gnutls_realloc_fast(list->node[hash].trusted_cas,
                                 (list->node[hash].trusted_ca_size +
@@ -192,6 +192,92 @@ gnutls_x509_trust_list_add_cas(gnutls_x509_trust_list_t list,
     return i;
 }
 
+static gnutls_x509_crt_t crt_cpy(gnutls_x509_crt_t src)
+{
+gnutls_x509_crt_t dst;
+int ret;
+
+	ret = gnutls_x509_crt_init(&dst);
+	if (ret < 0) {
+		gnutls_assert();
+		return NULL;
+	}
+	
+	ret = _gnutls_x509_crt_cpy(dst, src);
+	if (ret < 0) {
+		gnutls_assert();
+		return NULL;
+	}
+	
+	return dst;
+}
+
+/**
+ * gnutls_x509_trust_list_remove_cas:
+ * @list: The structure of the list
+ * @clist: A list of CAs
+ * @clist_size: The length of the CA list
+ *
+ * This function will remove the given certificate authorities
+ * from the trusted list.
+ *
+ * Note that this function can accept certificates and authorities
+ * not yet known. In that case they will be kept in a separate
+ * black list that will be used during certificate verification.
+ * Unlike gnutls_x509_trust_list_add_cas() there is no deinitialization
+ * restriction for  certificate list provided in this function.
+ *
+ * Returns: The number of removed elements is returned.
+ *
+ * Since: 3.1.10
+ **/
+int
+gnutls_x509_trust_list_remove_cas(gnutls_x509_trust_list_t list,
+                               const gnutls_x509_crt_t * clist,
+                               int clist_size)
+{
+    int i, r = 0;
+    unsigned j;
+    uint32_t hash;
+
+    for (i = 0; i < clist_size; i++) 
+      {
+        hash = hash_pjw_bare(clist[i]->raw_dn.data, clist[i]->raw_dn.size);
+        hash %= list->size;
+        
+        for (j=0;j<list->node[hash].trusted_ca_size;j++) 
+          {
+	    if (_gnutls_check_if_same_cert(clist[i], list->node[hash].trusted_cas[j]) != 0)
+	      {
+                gnutls_x509_crt_deinit(list->node[hash].trusted_cas[j]);
+                list->node[hash].trusted_cas[j] = 
+			list->node[hash].trusted_cas[list->node[hash].trusted_ca_size-1];
+		list->node[hash].trusted_ca_size--;
+		r++;
+              }
+          }
+        
+        /* Add the CA (or plain) certificate to the black list as well.
+	 * This will prevent a subordinate CA from being valid, and 
+	 * ensure that a server certificate will also get rejected.
+	  */
+	list->blacklisted =
+		    gnutls_realloc_fast(list->blacklisted,
+				(list->blacklisted_size + 1) *
+				sizeof(list->blacklisted[0]));
+	if (list->blacklisted == NULL)
+		return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
+
+	list->blacklisted[list->blacklisted_size] = crt_cpy(clist[i]);
+	if (list->blacklisted[list->blacklisted_size] != NULL)
+		list->blacklisted_size++;
+  
+         
+      }
+
+    return r;
+}
+
 /**
  * gnutls_x509_trust_list_add_named_crt:
  * @list: The structure of the list
@@ -223,24 +309,14 @@ gnutls_x509_trust_list_add_named_crt(gnutls_x509_trust_list_t list,
                                      const void *name, size_t name_size,
                                      unsigned int flags)
 {
-    gnutls_datum_t dn;
-    int ret;
     uint32_t hash;
 
     if (name_size >= MAX_SERVER_NAME_SIZE)
         return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
-    ret = gnutls_x509_crt_get_raw_issuer_dn(cert, &dn);
-    if (ret < 0) {
-        gnutls_assert();
-        return ret;
-    }
-
-    hash = hash_pjw_bare(dn.data, dn.size);
+    hash = hash_pjw_bare(cert->raw_issuer_dn.data, cert->raw_issuer_dn.size);
     hash %= list->size;
 
-    _gnutls_free_datum(&dn);
-
     list->node[hash].named_certs =
         gnutls_realloc_fast(list->node[hash].named_certs,
                             (list->node[hash].named_cert_size +
@@ -286,29 +362,19 @@ gnutls_x509_trust_list_add_crls(gnutls_x509_trust_list_t list,
                                 unsigned int verification_flags)
 {
     int ret, i, j = 0;
-    gnutls_datum_t dn;
     unsigned int vret = 0;
     uint32_t hash;
 
     /* Probably we can optimize things such as removing duplicates
      * etc.
      */
-
     if (crl_size == 0 || crl_list == NULL)
         return 0;
 
     for (i = 0; i < crl_size; i++) {
-        ret = gnutls_x509_crl_get_raw_issuer_dn(crl_list[i], &dn);
-        if (ret < 0) {
-            gnutls_assert();
-            return i;
-        }
-
-        hash = hash_pjw_bare(dn.data, dn.size);
+        hash = hash_pjw_bare(crl_list[i]->raw_issuer_dn.data, crl_list[i]->raw_issuer_dn.size);
         hash %= list->size;
 
-        _gnutls_free_datum(&dn);
-
         if (flags & GNUTLS_TL_VERIFY_CRL) {
 
             ret =
@@ -349,10 +415,8 @@ static int shorten_clist(gnutls_x509_trust_list_t list,
                          gnutls_x509_crt_t * certificate_list,
                          unsigned int clist_size)
 {
-    int ret;
     unsigned int j, i;
     uint32_t hash;
-    gnutls_datum_t dn;
 
     if (clist_size > 1) {
         /* Check if the last certificate in the path is self signed.
@@ -366,7 +430,7 @@ static int shorten_clist(gnutls_x509_trust_list_t list,
          */
         if (gnutls_x509_crt_check_issuer(certificate_list[clist_size - 1],
                                          certificate_list[clist_size -
-                                                          1]) > 0) {
+                                                          1]) != 0) {
             clist_size--;
         }
     }
@@ -377,21 +441,13 @@ static int shorten_clist(gnutls_x509_trust_list_t list,
      * self-signed E but already removed above), and we trust B, remove
      * B, C and D. */
     for (i = 1; i < clist_size; i++) {
-        ret = gnutls_x509_crt_get_raw_issuer_dn(certificate_list[i], &dn);
-        if (ret < 0) {
-            gnutls_assert();
-            return ret;
-        }
-
-        hash = hash_pjw_bare(dn.data, dn.size);
+        hash = hash_pjw_bare(certificate_list[i]->raw_issuer_dn.data, certificate_list[i]->raw_issuer_dn.size);
         hash %= list->size;
 
-        _gnutls_free_datum(&dn);
-
         for (j = 0; j < list->node[hash].trusted_ca_size; j++) {
-            if (check_if_same_cert
+            if (_gnutls_check_if_same_cert
                 (certificate_list[i],
-                 list->node[hash].trusted_cas[j]) == 0) {
+                 list->node[hash].trusted_cas[j]) != 0) {
                 /* cut the list at the point of first the trusted certificate */
                 clist_size = i + 1;
                 break;
@@ -437,7 +493,7 @@ static gnutls_x509_crt_t* sort_clist(gnutls_x509_crt_t sorted[DEFAULT_MAX_VERIFY
           if (i==j) continue;
         
           if (gnutls_x509_crt_check_issuer(clist[i],
-                                           clist[j])) 
+                                           clist[j]) != 0) 
             {
               issuer[i] = j;
               break;
@@ -487,27 +543,18 @@ int gnutls_x509_trust_list_get_issuer(gnutls_x509_trust_list_t list,
                                       gnutls_x509_crt_t * issuer,
                                       unsigned int flags)
 {
-    gnutls_datum_t dn;
     int ret;
     unsigned int i;
     uint32_t hash;
 
-    ret = gnutls_x509_crt_get_raw_issuer_dn(cert, &dn);
-    if (ret < 0) {
-        gnutls_assert();
-        return ret;
-    }
-
-    hash = hash_pjw_bare(dn.data, dn.size);
+    hash = hash_pjw_bare(cert->raw_issuer_dn.data, cert->raw_issuer_dn.size);
     hash %= list->size;
 
-    _gnutls_free_datum(&dn);
-
     for (i = 0; i < list->node[hash].trusted_ca_size; i++) {
         ret =
             gnutls_x509_crt_check_issuer(cert,
                                          list->node[hash].trusted_cas[i]);
-        if (ret > 0) {
+        if (ret != 0) {
             *issuer = list->node[hash].trusted_cas[i];
             return 0;
         }
@@ -516,6 +563,26 @@ int gnutls_x509_trust_list_get_issuer(gnutls_x509_trust_list_t list,
     return GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
 }
 
+static
+int check_if_in_blacklist(gnutls_x509_crt_t * cert_list, unsigned int cert_list_size,
+	gnutls_x509_crt_t * blacklist, unsigned int blacklist_size)
+{
+unsigned i, j;
+
+	if (blacklist_size == 0)
+		return 0;
+	
+	for (i=0;i<cert_list_size;i++) {
+		for (j=0;j<blacklist_size;j++) {
+			if (_gnutls_check_if_same_cert(cert_list[i], blacklist[j]) != 0) {
+				return 1;
+			}
+		}
+	}
+
+	return 0;
+}
+
 /**
  * gnutls_x509_trust_list_verify_crt:
  * @list: The structure of the list
@@ -544,7 +611,6 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
                                   unsigned int *verify,
                                   gnutls_verify_output_function func)
 {
-    gnutls_datum_t dn;
     int ret;
     unsigned int i;
     uint32_t hash;
@@ -560,18 +626,17 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
     if (cert_list_size <= 0)
         return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
 
-    ret =
-        gnutls_x509_crt_get_raw_issuer_dn(cert_list[cert_list_size - 1],
-                                          &dn);
-    if (ret < 0) {
-        gnutls_assert();
-        return ret;
-    }
-
-    hash = hash_pjw_bare(dn.data, dn.size);
+    hash = hash_pjw_bare(cert_list[cert_list_size - 1]->raw_issuer_dn.data,
+    	cert_list[cert_list_size - 1]->raw_issuer_dn.size);
     hash %= list->size;
 
-    _gnutls_free_datum(&dn);
+    ret = check_if_in_blacklist(cert_list, cert_list_size,
+		list->blacklisted, list->blacklisted_size);
+    if (ret != 0) {
+	*verify |= GNUTLS_CERT_REVOKED;
+	*verify |= GNUTLS_CERT_INVALID;
+	return 0;
+    }
 
     *verify = _gnutls_x509_verify_certificate(cert_list, cert_list_size,
                                               list->node[hash].trusted_cas,
@@ -579,6 +644,21 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
                                               trusted_ca_size, flags,
                                               func);
 
+#define LAST_DN cert_list[cert_list_size-1]->raw_dn
+#define LAST_IDN cert_list[cert_list_size-1]->raw_issuer_dn
+
+    if ((*verify) & GNUTLS_CERT_SIGNER_NOT_FOUND &&
+      (LAST_DN.size != LAST_IDN.size || memcmp(LAST_DN.data, LAST_IDN.data, LAST_IDN.size) != 0)) 
+      {
+        hash = hash_pjw_bare(cert_list[cert_list_size - 1]->raw_dn.data, cert_list[cert_list_size - 1]->raw_dn.size);
+        hash %= list->size;
+        *verify = _gnutls_x509_verify_certificate(cert_list, cert_list_size,
+                                              list->node[hash].trusted_cas,
+                                              list->node[hash].
+                                              trusted_ca_size, flags,
+                                              func);
+      }
+
     if (*verify != 0 || (flags & GNUTLS_VERIFY_DISABLE_CRL_CHECKS))
         return 0;
 
@@ -596,17 +676,9 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
     }
 
     for (i = 0; i < cert_list_size - 1; i++) {
-        ret = gnutls_x509_crt_get_raw_issuer_dn(cert_list[i], &dn);
-        if (ret < 0) {
-            gnutls_assert();
-            return ret;
-        }
-
-        hash = hash_pjw_bare(dn.data, dn.size);
+        hash = hash_pjw_bare(cert_list[i]->raw_issuer_dn.data, cert_list[i]->raw_issuer_dn.size);
         hash %= list->size;
 
-        _gnutls_free_datum(&dn);
-
         ret = _gnutls_x509_crt_check_revocation(cert_list[i],
                                                 list->node[hash].crls,
                                                 list->node[hash].crl_size,
@@ -619,6 +691,7 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
     }
 
     return 0;
+
 }
 
 /**
@@ -650,26 +723,25 @@ gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
                                         unsigned int *verify,
                                         gnutls_verify_output_function func)
 {
-    gnutls_datum_t dn;
     int ret;
     unsigned int i;
     uint32_t hash;
 
-    ret = gnutls_x509_crt_get_raw_issuer_dn(cert, &dn);
-    if (ret < 0) {
-        gnutls_assert();
-        return ret;
-    }
-
-    hash = hash_pjw_bare(dn.data, dn.size);
+    hash = hash_pjw_bare(cert->raw_issuer_dn.data, cert->raw_issuer_dn.size);
     hash %= list->size;
 
-    _gnutls_free_datum(&dn);
+    ret = check_if_in_blacklist(&cert, 1,
+		list->blacklisted, list->blacklisted_size);
+    if (ret != 0) {
+	*verify |= GNUTLS_CERT_REVOKED;
+	*verify |= GNUTLS_CERT_INVALID;
+	return 0;
+    }
 
     *verify = GNUTLS_CERT_INVALID | GNUTLS_CERT_SIGNER_NOT_FOUND;
 
     for (i = 0; i < list->node[hash].named_cert_size; i++) {
-        if (check_if_same_cert(cert, list->node[hash].named_certs[i].cert) == 0) {      /* check if name matches */
+        if (_gnutls_check_if_same_cert(cert, list->node[hash].named_certs[i].cert) != 0) {      /* check if name matches */
             if (list->node[hash].named_certs[i].name_size == name_size &&
                 memcmp(list->node[hash].named_certs[i].name, name,
                        name_size) == 0) {
@@ -698,40 +770,24 @@ gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
     return 0;
 }
 
-/* return 0 if @cert is in @list, 1 if not, or < 0 on error. */
+/* return 1 if @cert is in @list, 0 if not */
 int
 _gnutls_trustlist_inlist (gnutls_x509_trust_list_t list,
 			  gnutls_x509_crt_t cert)
 {
-  gnutls_datum_t dn;
   int ret;
   unsigned int i;
   uint32_t hash;
 
-  ret = gnutls_x509_crt_get_raw_dn (cert, &dn);
-  if (ret < 0)
-    {
-      gnutls_assert();
-      return ret;
-    }
-
-  hash = hash_pjw_bare(dn.data, dn.size);
+  hash = hash_pjw_bare(cert->raw_dn.data, cert->raw_dn.size);
   hash %= list->size;
 
-  _gnutls_free_datum (&dn);
-
   for (i = 0; i < list->node[hash].trusted_ca_size; i++)
     {
-      ret = check_if_same_cert (cert, list->node[hash].trusted_cas[i]);
-      if (ret < 0)
-	{
-	  gnutls_assert ();
-	  return ret;
-	}
-
-      if (ret == 0)
-	return 0;
+      ret = _gnutls_check_if_same_cert (cert, list->node[hash].trusted_cas[i]);
+      if (ret != 0)
+        return 1;
     }
 
-  return 1;
+  return 0;
 }
